@@ -1,7 +1,7 @@
 /*****************************************************************************
-*   $Id: eiffel.c,v 7.1 1998/12/07 01:38:00 darren Exp $
+*   $Id: eiffel.c,v 8.8 2000/03/27 05:03:23 darren Exp $
 *
-*   Copyright (c) 1998, Darren Hiebert
+*   Copyright (c) 1998-1999, Darren Hiebert
 *
 *   This source code is released for free distribution under the terms of the
 *   GNU General Public License.
@@ -48,14 +48,15 @@ typedef enum eException { ExceptionNone, ExceptionEOF } exception_t;
 typedef enum eKeywordId {
     KEYWORD_NONE,
     KEYWORD_alias, KEYWORD_all, KEYWORD_and, KEYWORD_as, KEYWORD_check,
-    KEYWORD_class, KEYWORD_creation, KEYWORD_Current, KEYWORD_debug,
-    KEYWORD_deferred, KEYWORD_do, KEYWORD_else, KEYWORD_elseif, KEYWORD_end,
-    KEYWORD_ensure, KEYWORD_expanded, KEYWORD_export, KEYWORD_external,
-    KEYWORD_false, KEYWORD_feature, KEYWORD_from, KEYWORD_frozen, KEYWORD_if,
-    KEYWORD_implies, KEYWORD_indexing, KEYWORD_infix, KEYWORD_inherit,
-    KEYWORD_inspect, KEYWORD_invariant, KEYWORD_is, KEYWORD_like,
-    KEYWORD_local, KEYWORD_loop, KEYWORD_not, KEYWORD_obsolete, KEYWORD_old,
-    KEYWORD_once, KEYWORD_or, KEYWORD_prefix, KEYWORD_redefine, KEYWORD_rename,
+    KEYWORD_class, KEYWORD_create, KEYWORD_creation, KEYWORD_Current,
+    KEYWORD_debug, KEYWORD_deferred, KEYWORD_do, KEYWORD_else,
+    KEYWORD_elseif, KEYWORD_end, KEYWORD_ensure, KEYWORD_expanded,
+    KEYWORD_export, KEYWORD_external, KEYWORD_false, KEYWORD_feature,
+    KEYWORD_from, KEYWORD_frozen, KEYWORD_if, KEYWORD_implies,
+    KEYWORD_indexing, KEYWORD_infix, KEYWORD_inherit, KEYWORD_inspect,
+    KEYWORD_invariant, KEYWORD_is, KEYWORD_like, KEYWORD_local,
+    KEYWORD_loop, KEYWORD_not, KEYWORD_obsolete, KEYWORD_old, KEYWORD_once,
+    KEYWORD_or, KEYWORD_prefix, KEYWORD_redefine, KEYWORD_rename,
     KEYWORD_require, KEYWORD_rescue, KEYWORD_Result, KEYWORD_retry,
     KEYWORD_select, KEYWORD_separate, KEYWORD_strip, KEYWORD_then,
     KEYWORD_true, KEYWORD_undefine, KEYWORD_unique, KEYWORD_until,
@@ -97,6 +98,7 @@ typedef enum eTagType {
     TAG_UNDEFINED,
     TAG_CLASS,
     TAG_FEATURE,
+    TAG_QUALIFIED_FEATURE,
     TAG_LOCAL,
     TAG_COUNT		/* must be last */
 } tagType;
@@ -124,6 +126,7 @@ static const keywordDesc EiffelKeywordTable[] = {
     { "as",		KEYWORD_as		},
     { "check",		KEYWORD_check		},
     { "class",		KEYWORD_class		},
+    { "create",		KEYWORD_create		},
     { "creation",	KEYWORD_creation	},
     { "current",	KEYWORD_Current		},
     { "debug",		KEYWORD_debug		},
@@ -154,7 +157,8 @@ static const keywordDesc EiffelKeywordTable[] = {
     { "not",		KEYWORD_not		},
     { "obsolete",	KEYWORD_obsolete	},
     { "old",		KEYWORD_old		},
-    { "once",		KEYWORD_or		},
+    { "once",		KEYWORD_once		},
+    { "or",		KEYWORD_or		},
     { "prefix",		KEYWORD_prefix		},
     { "redefine",	KEYWORD_redefine	},
     { "rename",		KEYWORD_rename		},
@@ -196,14 +200,19 @@ static void parseFreeOperator __ARGS((vString *const string, const int firstChar
 static keywordId analyzeToken __ARGS((vString *const name));
 static void readToken __ARGS((tokenInfo *const token));
 static boolean isIdentifierMatch __ARGS((const tokenInfo *const token, const char *const name));
-static void locateToken __ARGS((const tokenType type, const boolean skipPast, tokenInfo *const token));
-static void locateKeyword __ARGS((const keywordId keyword, const boolean skipPast, tokenInfo *const token));
+static void locateToken __ARGS((tokenInfo *const token, const tokenType type));
 static void skipPastMatchingBracket __ARGS((tokenInfo *const token));
 static void skipEntityType __ARGS((tokenInfo *const token));
 static void parseLocal __ARGS((tokenInfo *const token));
 static void findFeatureEnd __ARGS((tokenInfo *const token));
+static boolean readFeatureName __ARGS((tokenInfo *const token));
 static void parseFeature __ARGS((tokenInfo *const token));
 static void parseExport __ARGS((tokenInfo *const token));
+static void locateKeyword __ARGS((tokenInfo *const token, const keywordId keyword));
+static void parseFeatureClauses __ARGS((tokenInfo *const token));
+static void parseRename __ARGS((tokenInfo *const token));
+static void parseInherit __ARGS((tokenInfo *const token));
+static void parseClass __ARGS((tokenInfo *const token));
 static tokenInfo *newToken __ARGS((void));
 static void deleteToken __ARGS((tokenInfo *const token));
 static void init __ARGS((void));
@@ -222,7 +231,7 @@ static void buildEiffelKeywordHash()
     {
 	const keywordDesc *p = &EiffelKeywordTable[i];
 
-	addKeyword(p->name, LANG_EIFFEL, (int)p->id); 
+	addKeyword(p->name, LANG_EIFFEL, (int)p->id);
     }
 }
 
@@ -236,7 +245,8 @@ static const char *tagName( type )
     /*  Note that the strings in this array must correspond to the types in
      *  the tagType enumeration.
      */
-    static const char *names[] = { "undefined", "class", "feature", "local" };
+    static const char *names[] = {
+	"undefined", "class", "feature", "feature", "local" };
 
     Assert(sizeof(names)/sizeof(names[0]) == TAG_COUNT);
 
@@ -249,7 +259,7 @@ static int tagLetter( type )
     /*  Note that the characters in this list must correspond to the types in
      *  the tagType enumeration.
      */
-    const char *const chars = "?cfl";
+    const char *const chars = "?cffl";
 
     Assert(strlen(chars) == TAG_COUNT);
 
@@ -260,16 +270,17 @@ static boolean includeTag( type, fileScope )
     const tagType type;
     const boolean fileScope;
 {
-    const struct sEiffelInclude *const inc = &Option.include.eiffel;
+    const struct sEiffelInclude *const inc = &Option.include.eiffelTypes;
     boolean include = FALSE;
 
     if (fileScope  &&  ! Option.include.fileScope)
 	include = FALSE;
     else switch (type)
     {
-	case TAG_CLASS:		include = inc->classNames;	break;
-	case TAG_FEATURE:	include = inc->features;	break;
-	case TAG_LOCAL:		include = inc->localEntities;	break;
+	case TAG_CLASS:			include = inc->classNames;	break;
+	case TAG_FEATURE:		include = inc->features;	break;
+	case TAG_QUALIFIED_FEATURE:	include = inc->classPrefix;	break;
+	case TAG_LOCAL:			include = inc->localEntities;	break;
 
 	default: Assert("Bad type in Eiffel file" == NULL); break;
     }
@@ -286,10 +297,8 @@ static void qualifyClassTag( token )
 
 	initTagEntry(&e, name);
 
-	e.ctagsTruncate	= TRUE;
-	e.etagsIncName	= TRUE;
-	e.kindName	= tagName(TAG_CLASS);
-	e.kind		= tagLetter(TAG_CLASS);
+	e.kindName = tagName(TAG_CLASS);
+	e.kind     = tagLetter(TAG_CLASS);
 
 	makeTagEntry(&e);
     }
@@ -320,6 +329,16 @@ static void qualifyFeatureTag( token )
 	e.otherFields.value	= otherFields[1];
 
 	makeTagEntry(&e);
+
+	if (includeTag(TAG_QUALIFIED_FEATURE, isFileScope))
+	{
+	    vString* qualified = vStringNewInit(vStringValue(token->className));
+	    vStringPut(qualified, '.');
+	    vStringCat(qualified, token->string);
+	    e.name = vStringValue(qualified);
+	    makeTagEntry(&e);
+	    vStringDelete(qualified);
+	}
     }
     vStringCopy(token->featureName, token->string);
 }
@@ -443,7 +462,11 @@ static int parseEscapedCharacter()
 	case 'H':	d = '\\';	break;
 	case 'L':	d = '~';	break;
 	case 'N':	d = '\n';	break;
+#ifdef QDOS
+	case 'Q':	d = 0x9F;	break;
+#else
 	case 'Q':	d = '`';	break;
+#endif
 	case 'R':	d = '\r';	break;
 	case 'S':	d = '#';	break;
 	case 'T':	d = '\t';	break;
@@ -578,7 +601,7 @@ getNextChar:
 	case ')':	token->type = TOKEN_CLOSE_PAREN;	break;
 	case ',':	token->type = TOKEN_COMMA;		break;
 	case '.':	token->type = TOKEN_DOT;		break;
-	case ';':	token->type = TOKEN_SEPARATOR;		break;
+	case ';':	goto getNextChar;
 	case '[':	token->type = TOKEN_OPEN_BRACKET;	break;
 	case ']':	token->type = TOKEN_CLOSE_BRACKET;	break;
 	case '{':	token->type = TOKEN_OPEN_BRACE;		break;
@@ -691,30 +714,15 @@ static boolean isIdentifierMatch( token, name )
 		     strequiv(vStringValue(token->string), name));
 }
 
-static void locateToken( type, skipPast, token )
-    const tokenType type;
-    const boolean skipPast;
+static void locateToken( token, type )
     tokenInfo *const token;
+    const tokenType type;
 {
     do
 	readToken(token);
     while (token->type != type);
 
-    if (skipPast)
-	readToken(token);
-}
-
-static void locateKeyword( keyword, skipPast, token )
-    const keywordId keyword;
-    const boolean skipPast;
-    tokenInfo *const token;
-{
-    do
-	readToken(token);
-    while (! isKeyword(token, keyword));
-
-    if (skipPast)
-	readToken(token);
+    readToken(token);
 }
 
 static void skipPastMatchingBracket( token )
@@ -739,10 +747,16 @@ static void skipEntityType( token )
      */
     readToken(token);
 
+    if (isKeyword(token, KEYWORD_expanded))
+	readToken(token);
+
     /*  Skip over the type name, with possible generic parameters.
      */
-    if (isKeyword(token, KEYWORD_like))
+    if (isKeyword(token, KEYWORD_like)  ||
+	strequiv(vStringValue(token->string), "BIT"))
+    {
 	readToken(token);
+    }
     readToken(token);
     if (isType(token, TOKEN_OPEN_BRACKET))
 	skipPastMatchingBracket(token);
@@ -752,7 +766,11 @@ static void parseLocal( token )
     tokenInfo *const token;
 {
     readToken(token);	    /* skip past "local" keyword */
-    do
+
+    /*  Check keyword first in case local clause is empty
+     */
+    while (! isKeyword(token, KEYWORD_do)  &&
+	   ! isKeyword(token, KEYWORD_once))
     {
 	if (isType(token, TOKEN_IDENTIFIER))
 	    qualifyLocalTag (token);
@@ -760,8 +778,7 @@ static void parseLocal( token )
 	readToken(token);
 	if (isType(token, TOKEN_COLON))
 	    skipEntityType(token);
-    } while (! isKeyword(token, KEYWORD_do)  &&
-	     ! isKeyword(token, KEYWORD_once));
+    }
 }
 
 static void findFeatureEnd( token )
@@ -774,8 +791,11 @@ static void findFeatureEnd( token )
 	default:
 	    break;
 
+	case KEYWORD_deferred:
 	case KEYWORD_do:
+	case KEYWORD_external:
 	case KEYWORD_local:
+	case KEYWORD_obsolete:
 	case KEYWORD_once:
 	case KEYWORD_require:
 	{
@@ -811,36 +831,41 @@ static void findFeatureEnd( token )
     }
 }
 
+static boolean readFeatureName( token )
+    tokenInfo *const token;
+{
+    boolean isFeatureName = FALSE;
+
+    if (isType(token, TOKEN_IDENTIFIER))
+	isFeatureName = TRUE;
+    else if (isKeyword(token, KEYWORD_infix)  ||
+	    isKeyword(token, KEYWORD_prefix))
+    {
+	readToken(token);
+	if (isType(token, TOKEN_STRING))
+	    isFeatureName = TRUE;
+    }
+    return isFeatureName;
+}
+
 static void parseFeature( token )
     tokenInfo *const token;
 {
     do
     {
-	/*  Generate tag for either the feature identifier, or the string
-	 *  following either prefix or infix.
-	 */
-	if (isType(token, TOKEN_IDENTIFIER))
+	if (readFeatureName(token))
 	    qualifyFeatureTag(token);
-	else if (isKeyword(token, KEYWORD_infix)  ||
-		isKeyword(token, KEYWORD_prefix))
-	{
-	    readToken(token);
-	    if (isType(token, TOKEN_STRING))
-		qualifyFeatureTag(token);
-	}
 	readToken(token);
     } while (isType(token, TOKEN_COMMA));
 
 
-    /*  Skip over arguments and return type.
+    /*  Skip signature.
      */
-    if (isType(token, TOKEN_OPEN_PAREN))
-	locateToken(TOKEN_CLOSE_PAREN, TRUE, token);
-    if (isType(token, TOKEN_COLON))
+    if (isType(token, TOKEN_OPEN_PAREN))	/* arguments? */
+	locateToken(token, TOKEN_CLOSE_PAREN);
+    if (isType(token, TOKEN_COLON))		/* a query? */
 	skipEntityType(token);
 
-    /*  Skip over body if one exists.
-     */
     if (isKeyword(token, KEYWORD_is))
 	findFeatureEnd(token);
 }
@@ -851,7 +876,7 @@ static void parseExport( token )
     token->isExported = TRUE;
 
     readToken(token);
-    if (isType(token, TOKEN_OPEN_BRACE))
+    if (isType(token, TOKEN_OPEN_BRACE)) do
     {
 	readToken(token);
 	if (isType(token, TOKEN_CLOSE_BRACE))
@@ -863,9 +888,106 @@ static void parseExport( token )
 	{
 	    if (isIdentifierMatch(token, "NONE"))
 		token->isExported = FALSE;
-	    locateToken(TOKEN_CLOSE_BRACE, TRUE, token);
+	    locateToken(token, TOKEN_CLOSE_BRACE);
 	}
+    } while (isType(token, TOKEN_COMMA));
+}
+
+static void locateKeyword( token, keyword )
+    tokenInfo *const token;
+    const keywordId keyword;
+{
+    do
+	readToken(token);
+    while (! isKeyword(token, keyword));
+
+    readToken(token);
+}
+
+static void parseFeatureClauses( token )
+    tokenInfo *const token;
+{
+    do
+    {
+	while (isKeyword(token, KEYWORD_feature))
+	    parseExport(token);
+	parseFeature(token);
+    } while (! isKeyword(token, KEYWORD_end) &&
+	     ! isKeyword(token, KEYWORD_invariant));
+}
+
+static void parseRename( token )
+    tokenInfo *const token;
+{
+    do {
+	readToken(token);
+	if (readFeatureName(token))
+	{
+	    readToken(token);
+	    if (isKeyword(token, KEYWORD_as))
+	    {
+		readToken(token);
+		if (readFeatureName(token))
+		{
+		    qualifyFeatureTag(token);	    /* renamed feature */
+		    readToken(token);
+		}
+	    }
+	}
+    } while (isType(token, TOKEN_COMMA));
+
+    if (! isKeyword(token, KEYWORD_end))
+	locateKeyword(token, KEYWORD_end);
+}
+
+static void parseInherit( token )
+    tokenInfo *const token;
+{
+    do
+    {
+	readToken(token);
+	switch (token->keyword)	    /* check for feature adaptation */
+	{
+	    case KEYWORD_rename:
+		parseRename(token);
+		if (isKeyword(token, KEYWORD_end))
+		    readToken(token);
+		break;
+
+	    case KEYWORD_export:
+	    case KEYWORD_undefine:
+	    case KEYWORD_redefine:
+	    case KEYWORD_select:
+		locateKeyword(token, KEYWORD_end);
+		break;
+
+	    case KEYWORD_end:
+		readToken(token);
+		break;
+
+	    default: break;
+	}
+    } while (! isType(token, TOKEN_KEYWORD));
+}
+
+static void parseClass( token )
+    tokenInfo *const token;
+{
+    if (isType(token, TOKEN_IDENTIFIER))
+    {
+	qualifyClassTag(token);
+	readToken(token);
     }
+
+    do
+    {
+	switch (token->keyword)
+	{
+	    case KEYWORD_inherit:  parseInherit(token);        break;
+	    case KEYWORD_feature:  parseFeatureClauses(token); break;
+	    default:               readToken(token);           break;
+	}
+    } while (! isKeyword(token, KEYWORD_end));
 }
 
 static tokenInfo *newToken()
@@ -890,7 +1012,7 @@ static void deleteToken ( token )
     vStringDelete(token->className);
     vStringDelete(token->featureName);
 
-    free(token);
+    eFree(token);
 }
 
 static void init()
@@ -911,22 +1033,10 @@ extern boolean createEiffelTags()
 
     init();
     exception = (exception_t)(setjmp(Exception));
-    if (exception == ExceptionNone)
+    while (exception == ExceptionNone)
     {
-	locateKeyword(KEYWORD_class, TRUE, token);
-
-	if (isType(token, TOKEN_IDENTIFIER))
-	    qualifyClassTag(token);
-
-	locateKeyword(KEYWORD_feature, FALSE, token);
-	do
-	{
-	    if (isKeyword(token, KEYWORD_invariant))
-		locateKeyword(KEYWORD_feature, FALSE, token);
-	    if (isKeyword(token, KEYWORD_feature))
-		parseExport(token);
-	    parseFeature(token);
-	} while (! isKeyword(token, KEYWORD_end));
+	locateKeyword(token, KEYWORD_class);
+	parseClass(token);
     }
     deleteToken(token);
 
