@@ -1,7 +1,7 @@
 /*****************************************************************************
-*   $Id: entry.c,v 7.1 1998/12/07 01:38:00 darren Exp $
+*   $Id: entry.c,v 8.19 2000/04/23 17:35:51 darren Exp $
 *
-*   Copyright (c) 1996-1998, Darren Hiebert
+*   Copyright (c) 1996-1999, Darren Hiebert
 *
 *   This source code is released for free distribution under the terms of the
 *   GNU General Public License.
@@ -18,6 +18,9 @@
 #include <ctype.h>	/* to define isspace() */
 #include <errno.h>
 
+#if defined(HAVE_SYS_TYPES_H)
+# include <sys/types.h>	    /* to declare off_t on some hosts */
+#endif
 #if defined(HAVE_UNISTD_H)
 # include <unistd.h>	/* to declare close(), ftruncate(), truncate() */
 #endif
@@ -35,6 +38,8 @@
 #include "main.h"
 #include "options.h"
 #include "read.h"
+#include "sort.h"
+#include "strlist.h"
 
 /*============================================================================
 *=	Defines
@@ -68,10 +73,12 @@ tagFile TagFile = {
     NULL,		/* file pointer */
     { 0, 0 },		/* numTags */
     { 0, 0, 0 },	/* max */
-    { "", NULL },	/* etags */
+    { NULL, NULL, 0 },	/* etags */
     NULL		/* vLine */
 };
 
+static boolean TagsToStdout = FALSE;
+
 /*============================================================================
 =   Function prototypes
 ============================================================================*/
@@ -84,13 +91,19 @@ extern int ftruncate __ARGS((int fd, off_t length));
 #endif
 
 static void writePseudoTag __ARGS((const char *const tagName, const char *const fileName, const char *const pattern));
-static void updateSortedFlag __ARGS((const char *const line, FILE *const fp, const fpos_t *const pStartOfLine));
+static void updateSortedFlag __ARGS((const char *const line, FILE *const fp, fpos_t startOfLine));
 static void addPseudoTags __ARGS((void));
 static long unsigned int updatePseudoTags __ARGS((FILE *const fp));
 static boolean isValidTagAddress __ARGS((const char *const tag));
 static boolean isCtagsLine __ARGS((const char *const tag));
 static boolean isEtagsLine __ARGS((const char *const tag));
 static boolean isTagFile __ARGS((const char *const filename));
+#ifdef ENABLE_UPDATE
+static boolean isNameInList __ARGS((Arguments* const args, const char *const name, const size_t length));
+#endif
+static void sortTagFile __ARGS((void));
+static void resizeTagFile __ARGS((const long newSize));
+static void writeEtagsIncludes __ARGS((FILE *const fp));
 static size_t writeSourceLine __ARGS((FILE *const fp, const char *const line));
 static size_t writeCompactSourceLine __ARGS((FILE *const fp, const char *const line));
 static void rememberMaxLengths __ARGS((const size_t nameLength, const size_t lineLength));
@@ -106,7 +119,12 @@ static int writeCtagsEntry __ARGS((const tagEntryInfo *const tag));
 =   Function definitions
 ============================================================================*/
 
-extern const char *const tagFileName()
+extern void freeTagFileResources()
+{
+    vStringDelete(TagFile.vLine);
+}
+
+extern const char *tagFileName()
 {
     return TagFile.name;
 }
@@ -152,10 +170,10 @@ static void addPseudoTags()
     }
 }
 
-static void updateSortedFlag( line, fp, pStartOfLine )
+static void updateSortedFlag( line, fp, startOfLine )
     const char *const line;
     FILE *const fp;
-    const fpos_t *const pStartOfLine;
+    fpos_t startOfLine;
 {
     const char *const tab = strchr(line, '\t');
 
@@ -167,7 +185,7 @@ static void updateSortedFlag( line, fp, pStartOfLine )
 	{
 	    fpos_t nextLine;
 
-	    if (fgetpos(fp, &nextLine) == -1 || fsetpos(fp, pStartOfLine) == -1)
+	    if (fgetpos(fp, &nextLine) == -1 || fsetpos(fp, &startOfLine) == -1)
 		error(WARNING, "Failed to update 'sorted' pseudo-tag");
 	    else
 	    {
@@ -221,7 +239,7 @@ static unsigned long updatePseudoTags( fp )
 		tab == '\t')
 	    {
 		if (strcmp(classType, "_SORTED") == 0)
-		    updateSortedFlag(line, fp, &startOfLine);
+		    updateSortedFlag(line, fp, startOfLine);
 	    }
 	    fgetpos(fp, &startOfLine);
 	}
@@ -255,7 +273,7 @@ static boolean isValidTagAddress( excmd )
 	    if (sscanf(excmd, "%[^;\n]", address) == 1  &&
 		strspn(address,"0123456789") == strlen(address))
 		    isValid = TRUE;
-	    free(address);
+	    eFree(address);
 	}
     }
     return isValid;
@@ -296,7 +314,7 @@ static boolean isCtagsLine( line )
 	    isValidTagAddress(field(EXCMD)))
 		ok = TRUE;
 
-	free(fields);
+	eFree(fields);
     }
     return ok;
 }
@@ -332,27 +350,72 @@ static boolean isTagFile( filename )
     return ok;
 }
 
-extern void openTagFile( toStdout )
-    const boolean toStdout;
+extern void copyBytes( fromFp, toFp, size )
+    FILE* const fromFp;
+    FILE* const toFp;
+    const long size;
 {
-    static char tempName[L_tmpnam];
+    enum { BufferSize = 1000 };
+    long toRead, numRead;
+    char* buffer = (char*)eMalloc((size_t)BufferSize);
+    long remaining = size;
+    do
+    {
+	toRead = (0 < remaining && remaining < BufferSize) ?
+		    remaining : BufferSize;
+	numRead = fread(buffer, (size_t)1, (size_t)toRead, fromFp);
+	if (fwrite(buffer, (size_t)1, (size_t)numRead, toFp) < (size_t)numRead)
+	    error (FATAL | PERROR, "cannot complete write");
+	if (remaining > 0)
+	    remaining -= numRead;
+    } while (numRead == toRead  &&  remaining != 0);
+    eFree(buffer);
+}
+
+extern void copyFile( from, to, size )
+    const char *const from;
+    const char *const to;
+    const long size;
+{
+    FILE* const fromFp = fopen(from, "rb");
+    if (fromFp == NULL)
+	error (FATAL | PERROR, "cannot open file to copy");
+    else
+    {
+	FILE* const toFp = fopen(to, "wb");
+	if (toFp == NULL)
+	    error (FATAL | PERROR, "cannot open copy destination");
+	else
+	{
+	    copyBytes (fromFp, toFp, size);
+	    fclose(toFp);
+	}
+	fclose(fromFp);
+    }
+}
+
+extern void openTagFile()
+{
+    setDefaultTagFileName();
+    TagsToStdout = isDestinationStdout();
 
     if (TagFile.vLine == NULL)
 	TagFile.vLine = vStringNew();
 
     /*  Open the tags file.
      */
-    if (toStdout)
-    {
-	TagFile.name = tmpnam(tempName);
-	TagFile.fp = fopen(TagFile.name, "w");
-    }
+    if (TagsToStdout)
+	TagFile.fp = tempFile("w", &TagFile.name);
     else
     {
-	const char *const fname  = Option.tagFileName;
-	const boolean fileExists = doesFileExist(fname);
+	char* fname;
+	boolean fileExists;
 
+	setDefaultTagFileName();
+	fname = (char*)eMalloc(strlen(Option.tagFileName) + 1);
+	strcpy(fname, Option.tagFileName);
 	TagFile.name = fname;
+	fileExists = doesFileExist(fname);
 	if (fileExists  &&  ! isTagFile(fname))
 	    error(FATAL,
 	      "\"%s\" doesn't look like a tag file; I refuse to overwrite it.",
@@ -394,118 +457,119 @@ extern void openTagFile( toStdout )
 
 #ifdef USE_REPLACEMENT_TRUNCATE
 
-static int copyChars __ARGS((FILE *const fpFrom, FILE *const fpTo, const long size));
-static int copyFile __ARGS((const char *const from, const char *const to, const long size));
 static int replacementTruncate __ARGS((const char *const name, const long size));
 
-static int copyChars( fpFrom, fpTo, size )
-    FILE *const fpFrom;
-    FILE *const fpTo;
-    const long size;
-{
-    int result = -1;
-    long count;
-
-    for (count = 0  ;  count < size  ;  ++count)
-    {
-	const int c = getc(fpFrom);
-
-	if (c == EOF  ||  putc(c, fpTo) == EOF)
-	    break;
-    }
-    if (count == size)
-	result = 0;
-    return result;
-}
-
-static int copyFile( from, to, size )
-    const char *const from;
-    const char *const to;
-    const long size;
-{
-    int result = -1;
-    FILE *const fpFrom = fopen(from, "rb");
-
-    if (fpFrom != NULL)
-    {
-	FILE *const fpTo = fopen(to, "wb");
-
-	if (fpFrom != NULL)
-	{
-	    result = copyChars(fpFrom, fpTo, size);
-	    if (result == 0)
-		result = fclose(fpTo);
-	}
-	if (result == 0)
-	    result = fclose(fpFrom);
-    }
-    return result;
-}
-
 /*  Replacement for missing library function.
  */
 static int replacementTruncate( name, size )
     const char *const name;
     const long size;
 {
-    char tempName[L_tmpnam];
-    int result = -1;
+    char *tempName = NULL;
+    FILE *fp = tempFile("w", &tempName);
+    fclose(fp);
+    copyFile(name, tempName, size);
+    copyFile(tempName, name, WHOLE_FILE);
+    remove(tempName);
+    eFree(tempName);
 
-    if (tmpnam(tempName) != NULL)
-    {
-	result = copyFile(name, tempName, size);
-	if (result == 0)
-	    result = copyFile(tempName, name, size);
-	if (result == 0)
-	    result = remove(tempName);
-    }
-    return result;
+    return 0;
 }
 
 #endif
 
+static void sortTagFile()
+{
+    if (TagFile.numTags.added > 0L)
+    {
+	if (Option.sorted)
+	{
+	    if (Option.verbose)
+		printf("sorting tag file\n");
+#ifdef EXTERNAL_SORT
+	    externalSortTags(TagsToStdout);
+#else
+	    internalSortTags(TagsToStdout);
+#endif
+	}
+	else if (TagsToStdout)
+	    catFile(tagFileName());
+    }
+    if (TagsToStdout)
+	remove(tagFileName());		/* remove temporary file */
+}
+
+static void resizeTagFile( newSize )
+    const long newSize;
+{
+    int result = 0;
+
+#ifdef USE_REPLACEMENT_TRUNCATE
+    result = replacementTruncate(TagFile.name, newSize);
+#else
+# ifdef HAVE_TRUNCATE
+    result = truncate(TagFile.name, (off_t)newSize);
+# else
+    const int fd = open(TagFile.name, O_RDWR);
+
+    if (fd != -1)
+    {
+#  ifdef HAVE_FTRUNCATE
+	result = ftruncate(fd, (off_t)newSize);
+#  else
+#   ifdef HAVE_CHSIZE
+	result = chsize(fd, newSize);
+#   endif
+#  endif
+	close(fd);
+    }
+# endif
+#endif
+    if (result == -1)
+	fprintf(errout, "Cannot shorten tag file: errno = %d\n", errno);
+}
+
+static void writeEtagsIncludes( fp )
+    FILE *const fp;
+{
+    if (Option.etagsInclude)
+    {
+	unsigned int i;
+	for (i = 0  ;  i < stringListCount(Option.etagsInclude)  ;  ++i)
+	{
+	    vString *item = stringListItem(Option.etagsInclude, i);
+	    fprintf(fp, "\f\n%s,include\n", vStringValue(item));
+	}
+    }
+}
+
 extern void closeTagFile( resize )
     const boolean resize;
 {
-    const long __unused__ size = ftell(TagFile.fp);
+    long desiredSize, size;
 
+    if (Option.etags)
+	writeEtagsIncludes(TagFile.fp);
+    desiredSize = ftell(TagFile.fp);
+    fseek(TagFile.fp, 0L, SEEK_END);
+    size = ftell(TagFile.fp);
     fclose(TagFile.fp);
-    if (resize)
+    if (resize  &&  desiredSize < size)
     {
-	int __unused__ result = 0;
-
-#ifdef USE_REPLACEMENT_TRUNCATE
-	result = replacementTruncate(TagFile.name, size);
-#else
-# ifdef HAVE_TRUNCATE
-	result = truncate(TagFile.name, (off_t)size);
-# else
-	const int fd = open(TagFile.name, O_RDWR);
-
-	if (fd != -1)
-	{
-#  ifdef HAVE_FTRUNCATE
-	    result = ftruncate(fd, (off_t)size);
-#  else
-#   ifdef HAVE_CHSIZE
-	    result = chsize(fd, size);
-#   endif
-#  endif
-	    close(fd);
-	}
-# endif
-#endif
-	if (result == -1)
-	    fprintf(errout, "Cannot shorten tag file: errno = %d\n", errno);
+	DebugStatement(
+	    debugPrintf(DEBUG_STATUS, "shrinking %s from %ld to %ld bytes\n",
+			TagFile.name, size, desiredSize); )
+	resizeTagFile(desiredSize);
     }
+    if (! Option.etags)
+	sortTagFile();
+    eFree(TagFile.name);
+    TagFile.name = NULL;
 }
 
 extern void beginEtagsFile()
 {
-    tmpnam(TagFile.etags.name);
-    TagFile.etags.fp = fopen(TagFile.etags.name, "w+b");
-    if (TagFile.etags.fp == NULL)
-	error(FATAL | PERROR, "cannot open \"%s\"", TagFile.etags.name);
+    TagFile.etags.fp = tempFile("w+b", &TagFile.etags.name);
     TagFile.etags.byteCount = 0;
 }
 
@@ -522,8 +586,10 @@ extern void endEtagsFile( name )
 	    fputs(line, TagFile.fp);
 	fclose(TagFile.etags.fp);
 	remove(TagFile.etags.name);
+	eFree(TagFile.etags.name);
+	TagFile.etags.fp = NULL;
+	TagFile.etags.name = NULL;
     }
-    TagFile.etags.name[0] = '\0';
 }
 
 /*----------------------------------------------------------------------------
@@ -618,8 +684,8 @@ static void rememberMaxLengths( nameLength, lineLength )
 static int writeXrefEntry( tag )
     const tagEntryInfo *const tag;
 {
-    const char *const line = readSourceLine(TagFile.vLine, &tag->filePosition,
-					    NULL);
+    const char *const line =
+			readSourceLine(TagFile.vLine, tag->filePosition, NULL);
     int length = fprintf(TagFile.fp, "%-20s %-10s %4lu  %-14s ", tag->name,
 			 tag->kindName, tag->lineNumber, tag->sourceFileName);
 
@@ -660,17 +726,24 @@ static int writeEtagsEntry( tag )
     else
     {
 	long seekValue;
-	char *const line = readSourceLine(TagFile.vLine, &tag->filePosition,
-					  &seekValue);
+	char *const line =
+		readSourceLine(TagFile.vLine, tag->filePosition, &seekValue);
 
-	truncateTagLine(line, tag->name, TRUE);
-
-	if (tag->etagsIncName)
-	    length = fprintf(TagFile.etags.fp, "%s\177%s\001%lu,%ld\n", line,
-			     tag->name, tag->lineNumber, seekValue);
+	if (tag->truncateLine)
+	    truncateTagLine(line, tag->name, TRUE);
 	else
-	    length = fprintf(TagFile.etags.fp, "%s\177%lu,%ld\n", line,
-			     tag->lineNumber, seekValue);
+	{
+	    char *p;
+
+	    for (p = line  ;  *p != '\0'  ;  ++p)
+		if (strchr("\n\r", *p) != NULL)
+		{
+		    *p = '\0';
+		    break;
+		}
+	}
+	length = fprintf(TagFile.etags.fp, "%s\177%s\001%lu,%ld\n", line,
+			 tag->name, tag->lineNumber, seekValue);
     }
     TagFile.etags.byteCount += length;
 
@@ -705,7 +778,7 @@ static int addExtensionFlags( tag )
 	const char *const value = tag->otherFields.value[i];
 
 	if (label != NULL)
-	    length += fprintf(TagFile.fp, "%s%s: %s", separator,
+	    length += fprintf(TagFile.fp, "%s%s:%s", separator,
 			      label, value == NULL ? "" : value);
     }
 
@@ -715,12 +788,12 @@ static int addExtensionFlags( tag )
 static int writePatternEntry( tag )
     const tagEntryInfo *const tag;
 {
-    char *const line = readSourceLine(TagFile.vLine, &tag->filePosition, NULL);
+    char *const line = readSourceLine(TagFile.vLine, tag->filePosition, NULL);
     const int searchChar = Option.backward ? '?' : '/';
     boolean newlineTerminated;
     int length = 0;
 
-    if (tag->ctagsTruncate)
+    if (tag->truncateLine)
 	truncateTagLine(line, tag->name, FALSE);
     newlineTerminated = (boolean)(strchr("\n\r", line[strlen(line)-1]) != NULL);
 
@@ -786,8 +859,7 @@ extern void initTagEntry( e, name )
     e->filePosition	= getFilePosition();
     e->isFileScope	= FALSE;
     e->isFileEntry	= FALSE;
-    e->ctagsTruncate	= FALSE;
-    e->etagsIncName	= FALSE;
+    e->truncateLine	= FALSE;
     e->sourceFileName	= getFileName();
     e->name		= name;
     e->kindName		= "";
