@@ -629,11 +629,15 @@ getcmdline(firstc, count, indent)
 	case K_BS:
 	case Ctrl('H'):
 	case K_DEL:
+	case K_KDEL:
 	case Ctrl('W'):
 #ifdef FKMAP
 		if (cmd_fkmap && c == K_BS)
 		    c = K_DEL;
 #endif
+		if (c == K_KDEL)
+		    c = K_DEL;
+
 		/*
 		 * delete current character is the same as backspace on next
 		 * character, except at end of line
@@ -689,6 +693,7 @@ getcmdline(firstc, count, indent)
 		goto cmdline_changed;
 
 	case K_INS:
+	case K_KINS:
 #ifdef FKMAP
 		/* if Farsi mode set, we are in reverse insert mode -
 		   Do not change the mode */
@@ -721,7 +726,9 @@ getcmdline(firstc, count, indent)
 #endif
 		putcmdline('"');
 		++no_mapping;
-		c = safe_vgetc();
+		i = c = safe_vgetc(); /* CTRL-R <char> */
+		if (c == Ctrl('R'))
+		    c = safe_vgetc(); /* CTRL-R CTRL-R <char> */
 		--no_mapping;
 #ifdef WANT_EVAL
 		/*
@@ -749,7 +756,7 @@ getcmdline(firstc, count, indent)
 		}
 #endif
 		if (c != ESC)	    /* use ESC to cancel inserting register */
-		    cmdline_paste(c);
+		    cmdline_paste(c, i == Ctrl('R'));
 		redrawcmd();
 		goto cmdline_changed;
 
@@ -821,10 +828,10 @@ getcmdline(firstc, count, indent)
 			goto cmdline_not_changed;   /* Ignore mouse */
 # ifdef USE_GUI
 		if (gui.in_use)
-		    cmdline_paste('*');
+		    cmdline_paste('*', TRUE);
 		else
 # endif
-		    cmdline_paste(0);
+		    cmdline_paste(0, TRUE);
 		redrawcmd();
 		goto cmdline_changed;
 
@@ -1345,6 +1352,12 @@ set_cmdspos_cursor()
 	else
 #endif
 	    c = charsize(ccline.cmdbuff[i]);
+#ifdef MULTI_BYTE
+	/* multibyte wrap */
+	if (is_dbcs && IsLeadByte(ccline.cmdbuff[i])
+				 && (ccline.cmdspos % Columns + c == Columns))
+	    ccline.cmdspos++;
+#endif
 	/* If the cmdline doesn't fit, put cursor on last visible char. */
 	if ((ccline.cmdspos += c) >= m)
 	{
@@ -1474,7 +1487,8 @@ getexmodeline(c, dummy, indent)
 		    if (c1 == '\r')
 			c1 = '\n';
 
-		    if (c1 == BS || c1 == K_BS || c1 == DEL || c1 == K_DEL)
+		    if (c1 == BS || c1 == K_BS
+				  || c1 == DEL || c1 == K_DEL || c1 == K_KDEL)
 		    {
 			if (line_ga.ga_len > 0)
 			{
@@ -1679,6 +1693,7 @@ put_on_cmdline(str, len, redraw)
     int		retval;
     int		i;
     int		m;
+    int		c;
 
     if (len < 0)
 	len = STRLEN(str);
@@ -1727,19 +1742,25 @@ put_on_cmdline(str, len, redraw)
 		m = Columns * Rows;
 	    else
 		m = MAXCOL;
-	    while (len--)
+	    for (i = 0; i < len; ++i)
 	    {
 #ifdef CRYPTV
 		if (cmdline_crypt)
-		    i = 1;
+		    c = 1;
 		else
 #endif
-		    i = charsize(str[len]);
+		    c = charsize(str[i]);
+#ifdef MULTI_BYTE
+		/* multibyte wrap */
+		if (is_dbcs && IsLeadByte(str[i])
+				   && ccline.cmdspos % Columns + c == Columns)
+		    ccline.cmdspos++;
+#endif
 		/* Stop cursor at the end of the screen */
-		if (ccline.cmdspos + i >= m)
+		if (ccline.cmdspos + c >= m)
 		    break;
 		++ccline.cmdpos;
-		ccline.cmdspos += i;
+		ccline.cmdspos += c;
 	    }
 	}
     }
@@ -2129,15 +2150,16 @@ ExpandOne(str, orig, options, mode)
 	    if (options & WILD_HOME_REPLACE)
 		tilde_replace(str, cmd_numfiles, cmd_files);
 
-	    /*
-	     * Insert backslashes into a file name before a space, \, %, # and
-	     * wildmatch characters, except '~'.
-	     */
-	    if ((options & WILD_ESCAPE)
-		    && (expand_context == EXPAND_FILES
-			|| expand_context == EXPAND_BUFFERS
-			|| expand_context == EXPAND_DIRECTORIES))
+	    if (options & WILD_ESCAPE)
 	    {
+	      if (expand_context == EXPAND_FILES
+			|| expand_context == EXPAND_BUFFERS
+			|| expand_context == EXPAND_DIRECTORIES)
+	      {
+		/*
+		 * Insert a backslash into a file name before a space, \, %, #
+		 * and wildmatch characters, except '~'.
+		 */
 		for (i = 0; i < cmd_numfiles; ++i)
 		{
 		    /* for ":set path=" we need to escape spaces twice */
@@ -2175,8 +2197,40 @@ ExpandOne(str, orig, options, mode)
 			vim_free(cmd_files[i]);
 			cmd_files[i] = p;
 		    }
+
+		    /* If 'str' starts with "\~", replace "~" at start of
+		     * cmd_files[i] with "\~". */
+		    if (str[0] == '\\' && str[1] == '~'
+						    && cmd_files[i][0] == '~')
+		    {
+			p = alloc((unsigned)(STRLEN(cmd_files[i]) + 2));
+			if (p != NULL)
+			{
+			    p[0] = '\\';
+			    STRCPY(p + 1, cmd_files[i]);
+			    vim_free(cmd_files[i]);
+			    cmd_files[i] = p;
+			}
+		    }
 		}
 		expand_set_path = FALSE;
+	      }
+	      else if (expand_context == EXPAND_TAGS)
+	      {
+		/*
+		 * Insert a backslash before characters in a tag name that
+		 * would terminate the ":tag" command.
+		 */
+		for (i = 0; i < cmd_numfiles; ++i)
+		{
+		    p = vim_strsave_escaped(cmd_files[i], (char_u *)"\\|\"");
+		    if (p != NULL)
+		    {
+			vim_free(cmd_files[i]);
+			cmd_files[i] = p;
+		    }
+		}
+	      }
 	    }
 
 	    /*
@@ -3084,6 +3138,8 @@ get_histtype(name)
     return -1;
 }
 
+static int	last_maptick = -1;	/* last seen maptick */
+
 /*
  * Add the given string to the given history.  If the string is already in the
  * history then it is moved to the front.  "histype" may be HIST_CMD,
@@ -3095,7 +3151,6 @@ add_to_history(histype, new_entry, in_map)
     char_u	*new_entry;
     int		in_map;		/* consider maptick when inside a mapping */
 {
-    static int		last_maptick = -1;	/* last seen maptick */
     struct hist_entry	*hisptr;
 
     if (hislen == 0)		/* no history */
@@ -3301,9 +3356,15 @@ del_history_idx(histype, idx)
     i = calc_hist_idx(histype, idx);
     if (i < 0)
 	return FALSE;
-
     idx = hisidx[histype];
     vim_free(history[histype][i].hisstr);
+
+    /* When deleting the last added search string in a mapping, reset
+     * last_maptick, so that the last added search string isn't deleted again.
+     */
+    if (histype == HIST_SEARCH && maptick == last_maptick && i == idx)
+	last_maptick = -1;
+
     while (i != idx)
     {
 	j = (i + 1) % hislen;
