@@ -532,8 +532,9 @@ pixmap_create_by_dir(char *name, GdkPixmap **pixmap, GdkBitmap **mask)
 	STRCAT(full_pathname, name);
 	STRCAT(full_pathname, ".xpm");
 	if (mch_access((const char *)full_pathname, F_OK) == 0)
-	    *pixmap = gdk_pixmap_create_from_xpm(
-				    gui.mainwin->window,
+	    *pixmap = gdk_pixmap_colormap_create_from_xpm(
+				    NULL,
+				    gtk_widget_get_colormap(gui.mainwin),
 				    mask,
 				    &gui.mainwin->style->bg[GTK_STATE_NORMAL],
 				    (const char *)full_pathname);
@@ -944,29 +945,31 @@ gui_mch_dialog(int type,		/* type of dialog */
 	       char_u * buttons,	/* names of buttons */
 	       int def_but)		/* default button */
 {
-    char_u *names;
-    char_u *p;
-    int butcount;
-    int dialog_status = -1;
+    char_u		*names;
+    char_u		*p;
+    int			i;
+    int			butcount;
+    int			dialog_status = -1;
+    int			vertical;
 
-    GtkWidget *dialog;
-    GtkWidget *frame;
-    GtkWidget *vbox;
-    GtkWidget *table;
-    GtkWidget *pixmap;
-    GtkWidget *dialogmessage;
-    GtkWidget *action_area;
-    GtkWidget *sub_area;
-    GtkWidget *separator;
+    GtkWidget		*dialog;
+    GtkWidget		*frame;
+    GtkWidget		*vbox;
+    GtkWidget		*table;
+    GtkWidget		*pixmap;
+    GtkWidget		*dialogmessage;
+    GtkWidget		*action_area;
+    GtkWidget		*sub_area;
+    GtkWidget		*separator;
+    GtkAccelGroup	*accel_group;
 
-    GdkPixmap *icon = NULL;
-    GdkBitmap *mask = NULL;
-    char **icon_data = NULL;
+    GdkPixmap		*icon = NULL;
+    GdkBitmap		*mask = NULL;
+    char		**icon_data = NULL;
 
-#define MAXBUT 10
-    GtkWidget *button[MAXBUT];
-    ButtonData data[MAXBUT];
-    CancelData cancel_data;
+    GtkWidget		**button;
+    ButtonData		*data;
+    CancelData		cancel_data;
 
     if (title == NULL)
 	title = (char_u *) "Vim dialog...";
@@ -974,6 +977,9 @@ gui_mch_dialog(int type,		/* type of dialog */
     if ((type < 0) || (type > VIM_LAST_TYPE))
 	type = VIM_GENERIC;
 
+    /* Check 'v' flag in 'guioptions': vertical button placement. */
+    vertical = (vim_strchr(p_go, GO_VERTICAL) != NULL);
+
     /* if our pointer is currently hidden, then we should show it. */
     gui_mch_mousehide(FALSE);
 
@@ -1051,35 +1057,101 @@ gui_mch_dialog(int type,		/* type of dialog */
     gtk_box_pack_end(GTK_BOX(vbox), action_area, FALSE, TRUE, 0);
     gtk_widget_show(action_area);
 
-    sub_area = gtk_hbox_new(TRUE, 0);
+    /* Add a [vh]box in the hbox to center the buttons in the dialog. */
+    if (vertical)
+	sub_area = gtk_vbox_new(FALSE, 0);
+    else
+	sub_area = gtk_hbox_new(FALSE, 0);
     gtk_container_set_border_width(GTK_CONTAINER(sub_area), 0);
-    gtk_box_pack_start(GTK_BOX(action_area), sub_area, FALSE, TRUE, 0);
+    gtk_box_pack_start(GTK_BOX(action_area), sub_area, TRUE, FALSE, 0);
     gtk_widget_show(sub_area);
 
-    /* make a copy, so that we can insert NULs */
-    names = vim_strsave(buttons);
-    if (names == NULL)
-	return -1;
-
     /*
      * Create the buttons.
      */
-    p = names;
-    for (butcount = 0; butcount < MAXBUT; ++butcount) {
-	char_u *next;
 
+    /*
+     * Translate the Vim accelerator character into an underscore for GTK+.
+     * Double underscores to keep them in the label.
+     */
+    /* count the number of underscores */
+    i = 1;
+    for (p = buttons; *p; ++p)
+	if (*p == '_')
+	    ++i;
+
+    /* make a copy of "buttons" with the translated characters */
+    names = alloc(STRLEN(buttons) + i);
+    if (names == NULL)
+	return -1;
+
+    p = names;
+    for (i = 0; buttons[i]; ++i)
+    {
+	if (buttons[i] == DLG_HOTKEY_CHAR)
+	    *p++ = '_';
+	else
+	{
+	    if (buttons[i] == '_')
+		*p++ = '_';
+	    *p++ = buttons[i];
+	}
+    }
+    *p = NUL;
+
+    /* Count the number of buttons and allocate button[] and data[]. */
+    butcount = 1;
+    for (p = names; *p; ++p)
+	if (*p == DLG_BUTTON_SEP)
+	    ++butcount;
+    button = (GtkWidget **)alloc((unsigned)(butcount * sizeof(GtkWidget *)));
+    data = (ButtonData *)alloc((unsigned)(butcount * sizeof(ButtonData)));
+    if (button == NULL || data == NULL)
+    {
+	vim_free(names);
+	vim_free(button);
+	vim_free(data);
+	return -1;
+    }
+
+    /* Attach the new accelerator group to the window. */
+    accel_group = gtk_accel_group_new();
+    gtk_accel_group_attach(accel_group, GTK_OBJECT(dialog));
+
+    p = names;
+    for (butcount = 0; *p; ++butcount) {
+	char_u		*next;
+	GtkWidget	*label;
+	guint		accel_key;
+
+	/* Chunk out this single button. */
 	for (next = p; *next; ++next) {
-	    if (*next == DLG_HOTKEY_CHAR)
-		mch_memmove(next, next + 1, STRLEN(next));
 	    if (*next == DLG_BUTTON_SEP) {
 		*next++ = NUL;
 		break;
 	    }
 	}
 
-	button[butcount] = gtk_button_new_with_label((const gchar *)p);
+	button[butcount] = gtk_button_new();
 	GTK_WIDGET_SET_FLAGS(button[butcount], GTK_CAN_DEFAULT);
-	gtk_widget_show(button[butcount]);
+
+	label = gtk_accel_label_new("");
+        gtk_accel_label_set_accel_widget(GTK_ACCEL_LABEL(label), dialog);
+
+	accel_key = gtk_label_parse_uline(GTK_LABEL(label), (const gchar *)p);
+# ifdef GTK_USE_ACCEL
+	/* Don't add accelator if 'winaltkeys' is "no". */
+	if (accel_key != GDK_VoidSymbol) {
+	    gtk_widget_add_accelerator(button[butcount],
+		    "clicked",
+		    accel_group,
+		    accel_key, 0,
+		    0);
+	}
+# endif
+
+	gtk_container_add(GTK_CONTAINER(button[butcount]), label);
+	gtk_widget_show_all(button[butcount]);
 
 	data[butcount].status = &dialog_status;
 	data[butcount].index = butcount;
@@ -1089,18 +1161,14 @@ gui_mch_dialog(int type,		/* type of dialog */
 			   GTK_SIGNAL_FUNC(dlg_button_clicked),
 			   (gpointer) &data[butcount]);
 
-	if (*next == NUL) {
-	    gtk_box_pack_end(GTK_BOX(action_area), button[butcount],
-			     FALSE, TRUE, 0);
-	    break;
-	} else {
-	    gtk_box_pack_start(GTK_BOX(sub_area), button[butcount],
-			       TRUE, TRUE, 0);
-	}
+	gtk_box_pack_start(GTK_BOX(sub_area), button[butcount],
+			   TRUE, FALSE, 0);
 	p = next;
     }
+
     vim_free(names);
 
+    --butcount;
     --def_but;	    /* 1 is first button */
     if (def_but < 0)
 	def_but = 0;
@@ -1123,6 +1191,13 @@ gui_mch_dialog(int type,		/* type of dialog */
 
     if (dialog_status < 0)
 	dialog_status = 0;
+
+    /* let the garbage collector know that we don't need it anylonger */
+    gtk_accel_group_unref(accel_group);
+
+    vim_free(button);
+    vim_free(data);
+
     return dialog_status;
 }
 
