@@ -719,13 +719,22 @@ gui_x11_key_hit_cb(w, dud, event, dum)
     }
     if (len == 1 && (string[0] == Ctrl('C')
 #ifdef UNIX
-	    || string[0] == intr_char
+	    || (intr_char != 0 && string[0] == intr_char)
 #endif
 	    ))
     {
 	trash_input_buf();
 	got_int = TRUE;
     }
+
+    if (len == 1 && string[0] == CSI)
+    {
+	/* Turn CSI into K_CSI. */
+	string[1] = KS_EXTRA;
+	string[2] = KE_CSI;
+	len = 3;
+    }
+
     add_to_input_buf(string, len);
 
 #ifdef USE_XIM
@@ -1889,14 +1898,15 @@ gui_mch_beep()
 }
 
     void
-gui_mch_flash()
+gui_mch_flash(msec)
+    int		msec;
 {
     /* Do a visual beep by reversing the foreground and background colors */
     XFillRectangle(gui.dpy, gui.wid, gui.invert_gc, 0, 0,
 	    FILL_X((int)Columns) + gui.border_offset,
 	    FILL_Y((int)Rows) + gui.border_offset);
     XSync(gui.dpy, False);
-    ui_delay(20L, TRUE);	/* wait 1/50 of a second */
+    ui_delay((long)msec, TRUE);	/* wait for a few msec */
     XFillRectangle(gui.dpy, gui.wid, gui.invert_gc, 0, 0,
 	    FILL_X((int)Columns) + gui.border_offset,
 	    FILL_Y((int)Rows) + gui.border_offset);
@@ -2353,7 +2363,8 @@ gui_x11_wm_protocol_handler(w, client_data, event, dum)
 	return;
 
     /*
-     * SAVE_YOURSELF can arrive at any moment (not only when crashing!).
+     * The WM_SAVE_YOURSELF event arrives when the window manager wants to
+     * exit.  That can be cancelled though, thus Vim shouldn't exit here.
      * Just sync our swap files.
      */
     if (((XClientMessageEvent *)event)->data.l[0] ==
@@ -2361,6 +2372,11 @@ gui_x11_wm_protocol_handler(w, client_data, event, dum)
     {
 	out_flush();
 	ml_sync_all(FALSE, FALSE);	/* preserve all swap files */
+
+	/* Set the window's WM_COMMAND property, to let the window manager
+	 * know we are done saving ourselves.  We don't want to be restarted,
+	 * thus set argv to NULL. */
+	XSetCommand(gui.dpy, XtWindow(vimShell), NULL, 0);
 	return;
     }
 
