@@ -1452,12 +1452,12 @@ mch_dirname(buf, len)
 #endif
 }
 
-#ifdef __EMX__
+#if defined(__EMX__) || defined(PROTO)
 /*
  * Replace all slashes by backslashes.
  * When 'shellslash' set do it the other way around.
  */
-    static void
+    void
 slash_adjust(p)
     char_u  *p;
 {
@@ -1700,6 +1700,11 @@ mch_windexit(r)
 	    out_char('\n');
 	else
 	    msg_clr_eos();	/* clear the rest of the display */
+
+	/* Cursor may have been switched off without calling starttermcap()
+	 * when doing "vim -u vimrc" and vimrc contains ":q". */
+	if (full_screen)
+	    cursor_on();
     }
     out_flush();
     ml_close_all(TRUE);		/* remove all memfiles */
@@ -1852,7 +1857,7 @@ get_stty()
 	intr_char = keys.sg_kill;
 #endif
 	buf[1] = NUL;
-	add_termcode((char_u *)"kb", buf);
+	add_termcode((char_u *)"kb", buf, FALSE);
 
 	/*
 	 * If <BS> and <DEL> are now the same, redefine <DEL>.
@@ -1917,7 +1922,8 @@ check_mouse_termcode()
 # ifdef XTERM_MOUSE
     if (use_xterm_mouse())
     {
-	set_mouse_termcode(KS_MOUSE, (char_u *)"\033[M");
+	set_mouse_termcode(KS_MOUSE, (char_u *)(term_is_8bit(T_NAME)
+						       ? "\233M" : "\033[M"));
 	if (*p_mouse != NUL)
 	{
 	    /* force mouse off and maybe on to send possibly new mouse
@@ -2548,11 +2554,11 @@ mch_call_shell(cmd, options)
 			    if (ta_buf[i] == CSI && len - i > 2)
 			    {
 				c = TERMCAP2KEY(ta_buf[i + 1], ta_buf[i + 2]);
-				if (c == K_DEL || c == K_BS)
+				if (c == K_DEL || c == K_KDEL || c == K_BS)
 				{
 				    mch_memmove(ta_buf + i + 1, ta_buf + i + 3,
 						       (size_t)(len - i - 2));
-				    if (c == K_DEL)
+				    if (c == K_DEL || c == K_KDEL)
 					ta_buf[i] = DEL;
 				    else
 					ta_buf[i] = Ctrl('H');
@@ -4038,7 +4044,7 @@ do_xterm_trace()
 
     STRCPY(buf, mouse_code);
     strp = buf + STRLEN(buf);
-    *strp++ = xterm_button | MOUSE_DRAG;
+    *strp++ = (xterm_button | MOUSE_DRAG) & ~0x20;
     *strp++ = (char_u)(col + ' ' + 1);
     *strp++ = (char_u)(row + ' ' + 1);
     *strp = 0;
