@@ -38,15 +38,15 @@ ui_write(s, len)
 #endif
 }
 
-#if defined(USE_GUI) || defined(PROTO)
+#if (defined(USE_GUI) && defined(UNIX)) || defined(PROTO)
 /*
  * When executing an external program, there may be some typed characters that
  * are not consumed by it.  Give them back to ui_inchar() and they are stored
  * here for the next call.
  */
 static char_u *ta_str = NULL;
-static int ta_off;	    /* offset for next char to use */
-static int ta_len;	    /* length of ta_str */
+static int ta_off;	/* offset for next char to use when ta_str != NULL */
+static int ta_len;	/* length of ta_str when it's not NULL*/
 
     void
 ui_inchar_undo(s, len)
@@ -62,9 +62,14 @@ ui_inchar_undo(s, len)
     new = alloc(newlen);
     if (new != NULL)
     {
-	mch_memmove(new, ta_str + ta_off, (size_t)(ta_len - ta_off));
-	mch_memmove(new + ta_len - ta_off, s, (size_t)len);
-	vim_free(ta_str);
+	if (ta_str != NULL)
+	{
+	    mch_memmove(new, ta_str + ta_off, (size_t)(ta_len - ta_off));
+	    mch_memmove(new + ta_len - ta_off, s, (size_t)len);
+	    vim_free(ta_str);
+	}
+	else
+	    mch_memmove(new, s, (size_t)len);
 	ta_str = new;
 	ta_len = newlen;
 	ta_off = 0;
@@ -86,7 +91,7 @@ ui_inchar(buf, maxlen, wtime)
     int	    maxlen;
     long    wtime;	    /* don't use "time", MIPS cannot handle it */
 {
-#ifdef USE_GUI
+#if defined(USE_GUI) && defined(UNIX)
     /*
      * Use the typeahead if there is any.
      */
@@ -1288,6 +1293,7 @@ fill_input_buf(exit_on_error)
     inbufcount = 0;
 #  else
 
+    len = 0;	/* to avoid gcc warning */
     for (try = 0; try < 100; ++try)
     {
 	len = read(read_cmd_fd, (char *)inbuf + inbufcount,
@@ -1525,7 +1531,8 @@ clip_x11_request_selection(myShell, dpy)
 	XtGetSelectionValue(myShell, XA_PRIMARY, type,
 	    clip_x11_request_selection_cb, (XtPointer)&success, CurrentTime);
 
-	/* Do we need this?: */
+	/* Make sure the request for the selection goes out before waiting for
+	 * a response. */
 	XFlush(dpy);
 
 	/*
@@ -1538,7 +1545,7 @@ clip_x11_request_selection(myShell, dpy)
 	    if (XCheckTypedEvent(dpy, SelectionNotify, &event))
 		break;
 
-	    /* Do we need this?: */
+	    /* Do we need this?  Probably not. */
 	    XSync(dpy, False);
 	}
 
